/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Copyright (c) 2001  HASEGAWA Masafumi
 */

#include "ebi_bitmap.h"
#include "ebi_error.h"

__EBI_BEGIN_NAMESPACE

typedef struct
{
  const char* bytes;
  size_t length;
  size_t current;
}
BinaryInfo;

#pragma pack (1)
typedef struct
{
  UH bfType;
  UW bfSize;
  UH bfReserved1;
  UH bfReserved2;
  UW bfOffBits;
}
BitmapFileHeader;

typedef struct
{
  UW biSize;
  W biWidth;
  W biHeight;
  UH biPlanes;
  UH biBitCount;
  UW biCompression;
  UW biSizeImage;
  W biXPelsPerMeter;
  W biYPelsPerMeter;
  UW biClrUsed;
  UW biClrImportant;
}
BitmapInfoHeader;

typedef struct
{
  UB rgbBlue;
  UB rgbGreen;
  UB rgbRed;
  UB rgbReserved;
}
RgbQuad;
#pragma pack ()

typedef struct
{
  BitmapInfoHeader bmiHeader;
  RgbQuad bmiColors[1];
}
BitmapInfo;

Bitmap::Bitmap ()
{
  memset (&m_bmp, 0, sizeof (m_bmp));
  memset (&m_cspec, 0, sizeof (m_cspec));
}

Bitmap::~Bitmap ()
{
  clear ();
}

void Bitmap::clear ()
{
  if (m_bmp.baseaddr[0] != NULL)
    {
      free (m_bmp.baseaddr[0]);
    }
  memset (&m_bmp, 0, sizeof (m_bmp));

  if (m_cspec.colmap != NULL)
    {
      free (m_cspec.colmap);
    }
  memset (&m_cspec, 0, sizeof (m_cspec));
}

int Bitmap::getWidth () const
{
  return m_bmp.bounds.c.right;
}

int Bitmap::getHeight () const
{
  return m_bmp.bounds.c.bottom;
}

void Bitmap::draw (GID gid, W x, W y)
{
  ERR err = gset_bcs (gid, &m_cspec);
  if (err != E_OK)
    {
      Error::outputMessageBV ("gset_bcs", err);
    }

  RECT rect;
  rect = m_bmp.bounds;
  moverect (&rect, x, y);
  err = grst_bmp (gid, &rect, &m_bmp, &m_bmp.bounds, NULL, G_STORE | G_CVFORM | G_CVCOLOR);
  if (err != E_OK)
    {
      Error::outputMessageBV ("grst_bmp", err);
    }
}

static ERR readJpegBinary (UB* bytes, UW reqSize, VOID* source)
{
  BinaryInfo* binary = (BinaryInfo*) source;

  size_t remain = binary->length - binary->current;
  if (remain < reqSize)
    {
      reqSize = remain;
    }
  
  if (reqSize > 0)
    {
      memcpy (bytes, binary->bytes + binary->current, reqSize);
      binary->current += reqSize;
    }

  return reqSize;
}

bool Bitmap::convertJpeg (const char* bytes, int length)
{
  clear ();

  BinaryInfo binary;
  binary.bytes = bytes;
  binary.length = length;
  binary.current = 0;

  m_bmp.pixbits = 0x2018;
  ERR err = GetStdCSPEC (&m_cspec, 1, m_bmp.pixbits);
  if (err < 0)
    {
      Error::outputMessageBV ("GetStdCSPEC", err);
      return false;
    }

  IMG_BMP imgBmp;
  imgBmp.color_spec = &m_cspec;
  imgBmp.bmap = &m_bmp;
  imgBmp.funcptr = (ERR (*)()) readJpegBinary;
  imgBmp.flushptr = NULL;
  imgBmp.io_src = &binary;

  IMG_COMPACT imgCompact;
  imgCompact.method = LIBIMG_METHOD_JPEG;
  imgCompact.opt = LIBIMG_JPEG_READ_OPT_1_1_SIZE;

  err = libimg_rea_bmp (&imgBmp, &imgCompact);
  if (err < 0)
    {
      Error::outputMessageBV ("libimg_rea_bmp", err);
      return false;
    }

  return true;
}

bool Bitmap::convertBitmap (const char* binary, int length)
{
  clear ();

  BitmapFileHeader* bfh = (BitmapFileHeader*) binary;
  BitmapInfo* bi = (BitmapInfo*) (binary + sizeof (BitmapFileHeader));
  BitmapInfoHeader* bih = &bi->bmiHeader;
  RgbQuad* bmic = bi->bmiColors;

  // ѥåȥ顼Τб
  m_cspec.attr = 0x09;
  m_cspec.info[0] = bih->biClrUsed;
  m_cspec.colmap = (COLOR*) malloc (sizeof (COLOR) * m_cspec.info[0]);
  for (UW i = 0; i < bih->biClrUsed; ++i)
    {
      m_cspec.colmap[i] = 0x10000000 | (bmic[i].rgbRed << 16) | (bmic[i].rgbGreen << 8) | (bmic[i].rgbBlue);
    }

  m_bmp.planes = 1;
  m_bmp.pixbits = (bih->biBitCount << 8) | bih->biBitCount;
  m_bmp.rowbytes = (bih->biWidth + 15) / 16 * 2;
  m_bmp.bounds = toRECT (0, 0, bih->biWidth, bih->biHeight);

  W bytes = m_bmp.rowbytes * bih->biHeight;
  m_bmp.baseaddr[0] = (UB*) malloc (bytes);

  int srcRowbytes = (bih->biWidth + 7) / 8;
  UB *dest = m_bmp.baseaddr[0];
  const char* bits = binary + bfh->bfOffBits + srcRowbytes * (bih->biHeight - 1);
  for (int i = 0; i < bih->biHeight; ++i)
    {
      memcpy (dest, bits, srcRowbytes);
      dest[m_bmp.rowbytes - 1] = 0;
      bits -= srcRowbytes;
      dest += m_bmp.rowbytes;
    }

  return true;
}

__EBI_END_NAMESPACE
