/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Copyright (c) 2001  HASEGAWA Masafumi
 */

#include "ebi_device.h"
#include "ebi_bitmapfont.h"
#include "ebi_error.h"

__EBI_BEGIN_NAMESPACE

const COLOR Device::c_textColor = 0x10000000; // black
const COLOR Device::c_linkColor = 0x100000FF; // blue
const COLOR Device::c_backColor = 0x10FFFFFF; // white

Device::Device (GID gid)
{
  m_gid = gid;
  m_rect = toRECT (0, 0, 0, 0);

  m_fontSize = CHSSTD;
  m_fontAttr = 0;
  m_fontType = FONT_FULL;
  m_font = &m_fullFont;

  m_color = 0;
  memset (&m_halfFont, 0, sizeof (m_halfFont));
  memset (&m_fullFont, 0, sizeof (m_fullFont));
  updateFontSpec ();
}

Device::~Device ()
{
}

GID Device::getGid () const
{
  return m_gid;
}

void Device::updateFontSpec ()
{
  FSSPEC font;
  if (wget_inf (WI_FSWINDOW, &font, sizeof (font)) < 0)
    {
      memset (&m_fullFont, 0, sizeof (m_fullFont));
      m_fullFont.fclass = FTC_DEFAULT;
      m_fullFont.attr = m_fontAttr;
      m_fullFont.size = toSIZE (m_fontSize, m_fontSize);
    }
  else
    {
      m_fullFont = font;
      m_fullFont.size = toSIZE (m_fontSize, m_fontSize);
    }

  m_halfFont = m_fullFont;
  if (!(m_halfFont.attr & FT_PROP))
    {
      m_halfFont.size.h /= 2;
    }
}

W Device::getFontSize () const
{
  return m_fontSize;
}

void Device::setFontSize (W size)
{
  m_fontSize = size;

  updateFontSpec ();
}

UW Device::getFontAttr () const
{
  return m_fontAttr;
}

void Device::setFontAttr (UW attr)
{
  m_fontAttr = attr;

  updateFontSpec ();
}

void Device::selectFont (FontType fontType)
{
  if (fontType == m_fontType)
    {
      return;
    }

  switch (fontType)
    {
    case FONT_HALF:
      m_font = &m_halfFont;
      break;

    case FONT_FULL:
      m_font = &m_fullFont;
      break;
    }
  m_fontType = fontType;
  gset_fon (m_gid, m_font);
}

W Device::getCharWidth (TC code) const
{
  if (!(m_font->attr & FT_PROP))
    {
      return m_font->size.h;
    }

  return gget_chw (m_gid, code);
}

void Device::drawChar (TC code)
{
  gdra_chr (m_gid, code, G_STORE);
}

void Device::drawBitmapFont (BitmapFont* bitmapFont, const SIZE& size)
{
  W x, y;
  gget_chp (m_gid, &x, &y);

  if (bitmapFont != NULL)
    {
      RECT rect = toRECT (x, y + 1 - size.v, x + size.h, y + 1);
      bitmapFont->draw (m_gid, &rect, m_color, c_backColor);
    }

  gset_chp (m_gid, size.h, 0, False);
}

W Device::getMargin () const
{
  return m_fontSize / 2;
}

W Device::getLineFeed () const
{
  return m_fontSize * 3 / 2;
}

void Device::newPage ()
{
  m_color = c_textColor;
  ERR err = gset_chc (m_gid, m_color, c_backColor);
  if (err != E_OK)
    {
      Error::outputMessageBV ("gset_chc", err);
    }

  W x = getMargin ();
  W y = getMargin () + (getFontSize () - 1) - getLineFeed ();
  gset_chp (m_gid, x, y, True);
}

void Device::newLine ()
{
  W x, y;
  gget_chp (m_gid, &x, &y);
  x = getMargin ();
  y += getLineFeed ();
  gset_chp (m_gid, x, y, True);
}

void Device::beginLink ()
{
  m_color = c_linkColor;
  ERR err = gset_chc (m_gid, m_color, c_backColor);
  if (err != E_OK)
    {
      Error::outputMessageBV ("gset_chc", err);
    }
}

void Device::endLink ()
{
  m_color = c_textColor;
  ERR err = gset_chc (m_gid, m_color, c_backColor);
  if (err != E_OK)
    {
      Error::outputMessageBV ("gset_chc", err);
    }
}

W Device::calcLines () const
{
  W lines = (rectheight (m_rect) - getMargin ()) / getLineFeed ();
  return lines;
}

W Device::calcMax () const
{
  W max = (rectwidth (m_rect) - getMargin () * 2);
  return max;
}

PNT Device::calcPoint (const PNT& pnt, W low) const
{
  PNT p;
  p.x = pnt.x - getMargin ();
  p.y = (pnt.y - getMargin () / 2) / getLineFeed () + low;
  return p;
}

void Device::setRect (const RECT& rect)
{
  m_rect = rect;

  gset_vis (m_gid, m_rect);
}

void Device::clear ()
{
  PAT pat = { {0, 16, 16, c_backColor, 0, FILL100} };
  gfil_rec (m_gid, m_rect, &pat, 0, G_STORE);
}

__EBI_END_NAMESPACE


